#ifndef _CAVECAM
#define _CAVECAM

/*
 *
 * This is part of the ANL/MCS/FL CAVEav distribution.
 * written by Ivan R. Judson
 *
 */

#include <strings.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "CAVSharedObject.h"
#include "FLSharedMemory.h"
#include "FLImagePool.h"
#include "TexturedPlane.h"

typedef void (*fPtr)();

/** This is the CAVEcam base class from which all CAVEcam objects are
  derived.  The CAVEcam is an object that captures a frame of the CAVE
  scene from the location and direction specified by the user.  

  In order for the CAVcam to capture a frame of video it must render
  the scene from the CAVcam perspective. The CAVcam uses the
  gluLookAt() function call (p. 323 OpenGL Reference Manual) is is
  used to accomplish this.  
  */
  
class CAVcam : public CAVSharedObject {

public:
  /** @name  Methods */
  //@{

  /// The default constructor

  CAVcam(float ex = 0.0, float ey = 5.0, float ez = 0.0, 
	  float cx = 0.0, float cy = 5.0, float cz = -5.0,
	  float ux = 0.0, float uy = 1.0, float uz = 0.0);

  /// The default destructor

  virtual ~CAVcam();

  /// Accessor to set user drawing function

  void SetDrawFunc(void *df) { 
    DrawFunc = (fPtr)df; 
  }

  /// Accessor to set camera update function

  void SetUpdateFunc(void *df) {
    UpdateFunc = (fPtr)df;
  }

  void SetWidth(int _width) { 
    _fb_width = _width; 
  }

  void SetHeight(int _height) { 
    _fb_height = _height; 
  }
  
  void SetDepth(int _depth) { 
    _fb_depth = _depth; 
  }

  /// Accessor to set color of camera drawn, values are in the range [0.0, 1.0].

  void SetColor(float r, float g, float b) { 
    _red = r; _green = g; _blue = b; 
  }  

  /// Accessor to set the camera position in world coordinates.

  void SetPosition(float x, float y, float z) { 
    _eyex = x; _eyey = y; _eyez = z; 
  }

  /// Accessor sets vector specifying which direction is the front of the camera

  void SetFront(float x, float y, float z) { 
    _centerx = x; _centery = y; _centerz = z; 
  }

  /// Accessor sets vector specifying which direction is up for the camera

  void SetUp(float x, float y, float z) { 
    _upx = x; _upy = y; _upz = z; 
  }

  /// Accessor to set how often the camera should 'capture' frames from the CAVE scene.  The argument is the denominator of the fractional part of the frames to capture, e.g. if N = 4, then the camera will capture 1/4 of the frames.
  
  void SetFrameRate(int modFrames) { 
    _modFrames = modFrames;
  }

  void TurnOnCapture(void) {
    _captureFrames = 1;
  }
    
  void TurnOffCapture(void) {
    _captureFrames = 0;
  }

  int getCapture(void) {
    return _captureFrames;
  }
  
  void setFormat(GLenum f) {
    _format = f;
  }

  GLenum getFormat(void) {
    return _format;
  }

  void setFocalLength(int focalLength) {
    _focalLength = focalLength;
  }

  float getFocalLength(void) { return _focalLength; }

  /// Utility function to draw a 3d grid in CAVE space (-5, 0, -5) -> (+5, +10, +5)

  void Draw3DGrid(void);

  FLImageDescriptor *getCurrentFrame();

  void getFrameSize(int &width, int &height) {
    _imagepool->getFrameSize(width, height);
  }

  /// Update camera (call the UpdateFunction)
  
  void Update(void);
  
  /// Draw the camera

  void Draw(void);

  /// Render the frame

  void Process(void);

  /// Initialize the CAVcam
  
  virtual void Init(void);

  /// Store the Pixels

  virtual void PutPix(void);

    void setReflectImage(int r) { reflectImage_ = r; }

protected:

  /// This function calls the user-specified drawing function

  void CallDraw(void) {  
    (*DrawFunc)(); 
  }

  /// This function flips the frame since glReadPixels returns it upside down 

  void FlipFrame(int);

  /// Off screen rendering
  void setupOffscreen(void);

  int ComputeInverse4(float **a);

  void OpenGLToMatrix(float matrix[16], float **tomat)
    {
      int k = 0;
      for (int i = 0; i < 4; i++)
	{
	  for (int j = 0; j < 4; j++)
	    {
	      tomat[i][j] = matrix[k];
	      k++;
	    }
	}
    }
    
  // conversion functions since a [4][4] matrix is easier to deal with
       void MatrixToOpenGL(float **matrix, float tomat[16] )
    {
      int i, j;
      int k = 0;
      for (i = 0; i < 4; i++)
	{
	  for (j = 0; j < 4; j++)
	    {
	      tomat[k] = matrix[i][j];
	      k++;
	    }
	}
    }
  
  //@}

  /**@name Data */
  //@{

  /// Red Value of Camera's Color

  GLfloat _red;

  /// Green Value of Camera's Color

  GLfloat _green;

  /// Blue Value of Cameras Color

  GLfloat _blue;

  /// Camera X position

  GLdouble _eyex;

  /// Camera Y position

  GLdouble _eyey;

  /// Camera Z position

  GLdouble _eyez;

  /// Camera X direction

  GLdouble _centerx;

  /// Camera Y direction

  GLdouble _centery;

  /// Camera Z direction

  GLdouble _centerz;

  /// Camera X up

  GLdouble _upx;

  /// Camera Y up

  GLdouble _upy;

  /// Camera Z up

  GLdouble _upz;

  /// Camera X right

  GLdouble _rightx;

  /// Camera Y right

  GLdouble _righty;

  /// Camera Z right

  GLdouble _rightz;

  /// Inverse Viewing Matrix -- This is used to get the world coordinate position and orientation of the object that is drawn to represent the camera.  This is generated by taking the inverse of the viewing matrix that results from gluLookAt().
  
  GLfloat _invViewMat[16];

  /// Ring Buffer of image descriptors
	
  FLImagePool *_imagepool;

  /// Index of last image used from Ring Buffer

  int _lastimage;

  /// Serial number of last image used
  int _lastserial;

  /// Current image width

  int _fb_width;

  /// Current image height

  int _fb_height;

  /// Current image depth

  int _fb_depth;

  /// Current image size (height * width * depth)

  int _fb_size;

  /// Pointer to users drawing function
  fPtr DrawFunc;

  /// Pointer to the users camera location function
  fPtr UpdateFunc;

  /// Frame Counter
  int _fc;

  /// Do we capture frames or not?
  int _captureFrames;

  /// The denominator of the fractional part of the video stream to capture
  int _modFrames;

  /// The format of the pixel to read from the buffer
  GLenum _format;

  /// The focal length of the camera

  float _focalLength;

  /// Camera display

  Display *_camDisplay;

  /// Camera Visual

  XVisualInfo *_camVisual;

  /// This is the context the camera renders into

  GLXContext _camContext;

  /// The Pixmap to store the bits

  Pixmap pmap;

  /// the GLX Pixmap

  GLXPixmap glxpmap;

  /// View finder
  
  TexturedPlane *viewFinder_;

  /// Timestamp from last frame displayed
    
  double lastFrameTime_;
#ifdef PBUFFER
  /// Pbuffer CRAP
  GLXPbufferSGIX _pbuffer;
#endif

    int reflectImage_;
    static GLfloat reflectionMatrix_[4][4];
  //@}
};

#endif
