#include "JpegEncoder.h"
#include "Params.h"

#include <stdio.h>

JpegEncoder::JpegEncoder(int width, int height, DMimagepacking packing, int q)
{
    q_ = q;
    sourcePoolSize_ = 10;
    destPoolSize_ = 10;
    imageSize_ = width * height * 4;
    
    createConverter();
    configConverterSource(width, height, packing);
    configConverterDest();
    configConverter();
    configSourcePool();
    configDestPool();
}

JpegEncoder::~JpegEncoder()
{
    dmICDestroy(ic_);
    dmBufferDestroyPool(srcPool_);
    dmBufferDestroyPool(dstPool_);
}

void JpegEncoder::convertFrame(void *frame, DMbuffer *destBuffer)
{
     int selN;
     fd_set fds;
     FD_ZERO(&fds);
     FD_SET(srcPoolFd_, &fds);
     selN = select(srcPoolFd_ + 1, 0, &fds, 0, 0);

     DMbuffer buffer;
     if (dmBufferAllocate(srcPool_, &buffer) != DM_SUCCESS)
	 fatal("dmBufferAllocate failed");

     dmBufferSetSize(buffer, imageSize_);
     memcpy(dmBufferMapData(buffer), frame, imageSize_);

     if (dmICSend(ic_, buffer, 0, 0) != DM_SUCCESS)
	 fatal("dmICSend failed");
     dmBufferFree(buffer);

     //
     // Wait for it to come out
     //

     FD_ZERO(&fds);
     FD_SET(convFd_, &fds);
     selN = select(convFd_ + 1, &fds, 0, 0, 0);

     if (dmICReceive(ic_, destBuffer) != DM_SUCCESS)
	 fatal("dmICReceive failed");
}

void JpegEncoder::createConverter()
{
    int converter;

    converter = findJpegConverter(DM_IC_CODE_DIRECTION_ENCODE);

    if (dmICCreate(converter, &ic_) != DM_SUCCESS)
	fatal("dmICCreate failed");
}

void JpegEncoder::configConverterSource(int width, int height,
					DMimagepacking packing)
{
    Params params;

    if (dmSetImageDefaults(params.getDMparams(),
			   width, height, packing) != DM_SUCCESS)
	 fatal("dmSetImageDefaults failed");

     if (dmICSetSrcParams(ic_, params.getDMparams()) != DM_SUCCESS)
	 fatal("dmICSetSrcParams failed");
}

void JpegEncoder::configConverterDest()
{
    Params params;

    if (dmICGetDefaultDstParams(ic_, params.getDMparams()) != DM_SUCCESS)
	 fatal("dmICGetDstParams failed");

     printf("dst: size=%dx%d packing=%d\n",
	    params.getInt(DM_IMAGE_WIDTH),
	    params.getInt(DM_IMAGE_HEIGHT),
	    params.getInt(DM_IMAGE_PACKING));

     if (params.setInt(DM_JPEG_NOHEADERS, DM_TRUE) != DM_SUCCESS)
	fatal("noheaders failed");
	    
     if (dmICSetDstParams(ic_, params.getDMparams()) != DM_SUCCESS)
	 fatal("dmICSetSrcParams failed");
}

void JpegEncoder::configConverter()
{
    Params params;
    
    if(dmICGetDefaultConvParams(ic_, params.getDMparams()) != DM_SUCCESS)
	fatal("dmICGetDefaultConvParams failed");

    if (params.setString(DM_IMAGE_COMPRESSION, DM_IMAGE_JPEG)
	!= DM_SUCCESS)
	fatal("dmParamsSetString jpeg");

    if (params.setEnum(DM_IMAGE_ORIENTATION, DM_IMAGE_BOTTOM_TO_TOP)
	!= DM_SUCCESS)
	fatal("dmParamsSetEnum orient");

    double res = params.getFloat(DM_IMAGE_QUALITY_SPATIAL);
    printf("spatial res: %f\n", res);

    params.setFloat(DM_IMAGE_QUALITY_SPATIAL, (float) q_ / 100.0);

    res = params.getFloat(DM_IMAGE_QUALITY_SPATIAL);
    printf("spatial res: %f\n", res);

    if (dmICSetConvParams(ic_, params.getDMparams()) != DM_SUCCESS)
	fatal("dmICSetConvParams");

    convFd_ = dmICGetDstQueueFD(ic_);
    printf("got conv fd %d\n", convFd_);
}

void JpegEncoder::setQuality(int q)
{
    Params params;
    q_ = q;
    
    if(dmICGetConvParams(ic_, params.getDMparams()) != DM_SUCCESS)
	fatal("dmICGetConvParams failed");


    double res = params.getFloat(DM_IMAGE_QUALITY_SPATIAL);
    printf("spatial res: %f\n", res);

    params.setFloat(DM_IMAGE_QUALITY_SPATIAL, (float) q_ / 100.0);

    res = params.getFloat(DM_IMAGE_QUALITY_SPATIAL);
    printf("spatial res: %f\n", res);

    if (dmICSetConvParams(ic_, params.getDMparams()) != DM_SUCCESS)
	fatal("dmICSetConvParams");
}

void JpegEncoder::configSourcePool()
{
    Params params;

    if (dmBufferSetPoolDefaults(params.getDMparams(),
				sourcePoolSize_,
				imageSize_,
				DM_TRUE,
				DM_TRUE) != DM_SUCCESS)
	fatal("dmBufferSetPoolDefaults failed");

    if (dmICGetSrcPoolParams(ic_, params.getDMparams()) != DM_SUCCESS)
	fatal("dmICGetSrcPoolParams failed");

    if (dmBufferCreatePool(params.getDMparams(), &srcPool_) != DM_SUCCESS)
	 fatal("dmBufferCreatePool failed");
    
    if (dmBufferSetPoolSelectSize(srcPool_, imageSize_) != DM_SUCCESS)
	 fatal("dmBufferSetPoolSelectSize failed");

     srcPoolFd_ = dmBufferGetPoolFD(srcPool_);
}

void JpegEncoder::configDestPool()
{
    Params params;

    if (dmBufferSetPoolDefaults(params.getDMparams(),
				destPoolSize_,
				imageSize_,
				DM_TRUE,
				DM_TRUE) != DM_SUCCESS)
	fatal("dmBufferSetPoolDefaults failed");
    
    if (dmICGetDstPoolParams(ic_, params.getDMparams()) != DM_SUCCESS)
	fatal("dmICGetSrcPoolParams failed");

    if (dmBufferCreatePool(params.getDMparams(), &dstPool_) != DM_SUCCESS)
	fatal("dmBufferCreatePool failed");

    if (dmICSetDstPool(ic_, dstPool_) != DM_SUCCESS)
	fatal("dmICSetDstPool failed");
}

int JpegEncoder::findJpegConverter(DMcodedirection direction)
{
     int convRealtime = -1, convNonRealtime = -1;
     int converter;

     int n = dmICGetNum();
     printf("n=%d\n", n);
     
     while (n--)
     {
	 Params p;
         if (dmICGetDescription(n, p.getDMparams()) == DM_SUCCESS)
	 {
	     union {
		 char id[4];
		 int idi;
	     } idu;
	     idu.id[0] = 'j'; idu.id[1] = 'p';
	     idu.id[2] = 'e'; idu.id[3] = 'g';

	     if ( p.getInt(DM_IC_ID) == idu.idi)
	     {
		 int speed = p.getEnum(DM_IC_SPEED);
		 int thisDirection = p.getEnum(DM_IC_CODE_DIRECTION);

		 if (thisDirection != direction)
		     continue;
		 if (speed == DM_IC_SPEED_REALTIME)
		     convRealtime = n;
		 else if (speed == DM_IC_SPEED_NONREALTIME)
		     convNonRealtime = n;
	     }
	 }
     }
     printf("Got realtime converter %d nonrealtime %d\n",
	    convRealtime, convNonRealtime);

     if (convRealtime != -1)
	 converter = convRealtime;
     else
	 converter = convNonRealtime;

     return converter;
}

void JpegEncoder::fatal(char *msg)
{
    printf("Fatal error: %s\n", msg);

    int errn;
    const char *err = dmGetError(&errn, 0);
    printf("DM error %d: %s\n", errn, err);
    abort();
}
